"""
Version information for depthviz, created during installation.
"""

# These version placeholders will be replaced later during substitution.
__version__ = "0.5.1b1"
__version_tuple__ = (0, 5, "1b1")
